@@ -42,18 +42,24 @@
 (require 'jabber-core)
 (require 'jabber-alert)
 (require 'jabber-util)
+(require 'jabber-autoloads)
 (require 'cl)
 
 (defgroup jabber-activity nil
   "activity tracking options"
   :group 'jabber)
 
+;; All the (featurep 'jabber-activity) is so we don't call a function
+;; with an autoloaded cookie while the file is loading, since that
+;; would lead to endless load recursion.
+
 (defcustom jabber-activity-make-string 'jabber-activity-make-string-default
   "Function to call, for making the string to put in the mode
 line.  The default function returns the nick of the user."
   :set #'(lambda (var val)
 	   (custom-set-default var val)
-	   (when (fboundp 'jabber-activity-make-name-alist)
+	   (when (and (featurep 'jabber-activity)
+		      (fboundp 'jabber-activity-make-name-alist))
 	     (jabber-activity-make-name-alist)
 	     (jabber-activity-mode-line-update)))
   :type 'function
@@ -70,7 +76,8 @@ at least this long, when possible."
 JIDs."
   :set #'(lambda (var val)
 	   (custom-set-default var val)
-	   (when (fboundp 'jabber-activity-make-name-alist)
+	   (when (and (featurep 'jabber-activity)
+		      (fboundp 'jabber-activity-make-name-alist))
 	     (jabber-activity-make-name-alist)
 	     (jabber-activity-mode-line-update)))
   :type '(choice (function-item :tag "Keep strings"
@@ -86,7 +93,8 @@ JIDs."
   :group 'jabber-activity
   :set #'(lambda (var val)
 	   (custom-set-default var val)
-	   (when (bound-and-true-p jabber-activity-mode)
+	   (when (and (featurep 'jabber-activity)
+		      (bound-and-true-p jabber-activity-mode))
 	     (jabber-activity-mode -1)
 	     (jabber-activity-mode 1))))
 
@@ -97,7 +105,7 @@ Same syntax as `mode-line-format'."
   :type 'sexp
   :group 'jabber-activity
   :set #'(lambda (var val)
-	   (if (not (bound-and-true-p jabber-activity-mode))
+	   (if (not (and (featurep 'jabber-activity) (bound-and-true-p jabber-activity-mode)))
 	       (custom-set-default var val)
 	     (jabber-activity-mode -1)
 	     (custom-set-default var val)
@@ -267,13 +275,9 @@ on JIDs where `jabber-activity-show-p'"
 
 (defun jabber-activity-add (from buffer text proposed-alert)
   "Add a JID to mode line when `jabber-activity-show-p'"
-  ;; In case of private MUC message, we want to keep the full JID.
-  (let ((jid (if (jabber-muc-sender-p from)
-		 from
-	       (jabber-jid-user from))))
-    (when (funcall jabber-activity-show-p jid)
-      (add-to-list 'jabber-activity-jids jid)
-      (jabber-activity-mode-line-update))))
+  (when (funcall jabber-activity-show-p from)
+    (add-to-list 'jabber-activity-jids from)
+    (jabber-activity-mode-line-update)))
 
 (defun jabber-activity-add-muc (nick group buffer text proposed-alert)
   "Add a JID to mode line when `jabber-activity-show-p'"
@@ -281,6 +285,12 @@ on JIDs where `jabber-activity-show-p'"
     (add-to-list 'jabber-activity-jids group)
     (jabber-activity-mode-line-update)))
 
+(defun jabber-activity-presence (who oldstatus newstatus statustext proposed-alert)
+  "Add a JID to mode line on subscription requests."
+  (when (string= newstatus "subscribe")
+    (add-to-list 'jabber-activity-jids (symbol-name who))
+    (jabber-activity-mode-line-update)))
+
 (defun jabber-activity-kill-hook ()
   "Query the user as to whether killing Emacs should be cancelled
 when there are unread messages which otherwise would be lost, if
@@ -293,17 +303,27 @@ when there are unread messages which otherwise would be lost, if
 
 ;;; Interactive functions
 
+(defvar jabber-activity-last-buffer nil
+  "Last non-Jabber buffer used.")
+
 (defun jabber-activity-switch-to (&optional jid-param)
   "If JID-PARAM is provided, switch to that buffer.  If JID-PARAM is nil and
 there has been activity in another buffer, switch to that buffer.  If no such
-buffer exists, switch back to most recently used buffer."
-  (interactive)
-  (if (or jid-param jabber-activity-jids)
-      (let ((jid (or jid-param (car jabber-activity-jids))))
-	(switch-to-buffer (jabber-activity-find-buffer-name jid))
-	(jabber-activity-clean))
-    ;; Switch back to the buffer used last
-    (switch-to-buffer nil)))
+buffer exists, switch back to the last non Jabber chat buffer used."
+    (interactive)
+    (if (or jid-param jabber-activity-jids)
+        (let ((jid (or jid-param (car jabber-activity-jids))))
+          (unless (eq major-mode 'jabber-chat-mode)
+            (setq jabber-activity-last-buffer (current-buffer)))
+          (switch-to-buffer (jabber-activity-find-buffer-name jid))
+          (jabber-activity-clean))
+      (if (eq major-mode 'jabber-chat-mode)
+	  ;; Switch back to the buffer used last
+	  (when (buffer-live-p jabber-activity-last-buffer)
+	    (switch-to-buffer jabber-activity-last-buffer))
+	(message "No new activity"))))
+
+(defvar jabber-activity-idle-timer nil "Idle timer used for activity cleaning")
 
 ;;;###autoload
 (define-minor-mode jabber-activity-mode
@@ -325,8 +345,12 @@ With a numeric arg, enable this display if arg is positive."
 		  'jabber-activity-add)
 	(add-hook 'jabber-muc-hooks
 		  'jabber-activity-add-muc)
-	(add-hook 'jabber-post-connect-hook
-		  'jabber-activity-make-name-alist)
+	(add-hook 'jabber-presence-hooks
+		  'jabber-activity-presence)
+        (setq jabber-activity-idle-timer (run-with-idle-timer 2 t 'jabber-activity-clean))
+	;; XXX: reactivate
+	;; (add-hook 'jabber-post-connect-hooks
+;; 		  'jabber-activity-make-name-alist)
 	(add-to-list 'kill-emacs-query-functions
 		     'jabber-activity-kill-hook)
 	(add-to-list 'global-mode-string
@@ -357,8 +381,12 @@ With a numeric arg, enable this display if arg is positive."
 		   'jabber-activity-add)
       (remove-hook 'jabber-muc-hooks
 		   'jabber-activity-add-muc)
-      (remove-hook 'jabber-post-connect-hook
-		   'jabber-activity-make-name-alist)
+      (remove-hook 'jabber-presence-hooks
+		   'jabber-activity-presence)
+      (ignore-errors (cancel-timer jabber-activity-idle-timer))
+      ;; XXX: reactivate
+;;       (remove-hook 'jabber-post-connect-hooks
+;; 		   'jabber-activity-make-name-alist)
       (setq global-mode-string (delete '(t jabber-activity-mode-string)
 				       global-mode-string))
       (setq frame-title-format
