@@ -1,7 +1,7 @@
 ;; jabber-alert.el - alert hooks
 
+;; Copyright (C) 2003, 2004, 2005, 2007, 2008 - Magnus Henoch - mange@freemail.hu
 ;; Copyright (C) 2002, 2003, 2004 - tom berger - object@intelectronica.net
-;; Copyright (C) 2003, 2004, 2005 - Magnus Henoch - mange@freemail.hu
 
 ;; This file is a part of jabber.el.
 
@@ -20,6 +20,7 @@
 ;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 (require 'jabber-util)
+(require 'jabber-autoloads)
 
 (require 'cl)
 
@@ -106,8 +107,7 @@ every time."
   :group 'jabber-alerts)
 
 (defcustom jabber-alert-presence-hooks 
-  '(jabber-presence-update-roster
-    jabber-presence-echo)
+  '(jabber-presence-echo)
   "Hooks run when a user's presence changes.
 
 Arguments are WHO, OLDSTATUS, NEWSTATUS, STATUSTEXT and
@@ -120,7 +120,6 @@ one of \"subscribe\", \"unsubscribe\", \"subscribed\" and
   :type 'hook
   :options '(jabber-presence-beep
 	     jabber-presence-wave
-	     jabber-presence-update-roster
 	     jabber-presence-switch
 	     jabber-presence-display
 	     jabber-presence-echo)
@@ -221,6 +220,11 @@ files."
   :type 'file
   :group 'jabber-alerts)
 
+(defcustom jabber-play-sound-file 'play-sound-file
+  "a function to call to play alert sound files"
+  :type 'function
+  :group 'jabber-alerts)
+
 (defmacro define-jabber-alert (name docstring function)
   "Define a new family of external alert hooks.
 Use this macro when your hooks do nothing except displaying a string
@@ -272,13 +276,6 @@ Examples:
 (define-jabber-alert beep "Beep on event"
   (lambda (&rest ignore) (beep)))
 
-;; External notifiers
-(require 'jabber-screen)
-(require 'jabber-ratpoison)
-(require 'jabber-sawfish)
-(require 'jabber-festival)
-(require 'jabber-xmessage)
-
 ;; Message alert hooks
 (defun jabber-message-default-message (from buffer text)
   (when (or jabber-message-alert-same-buffer
@@ -309,7 +306,7 @@ Examples:
 			       (return (cdr entry))))
 			   jabber-alert-message-wave)))
       (unless (equal sound-file "")
-	(play-sound-file sound-file)))))
+	(funcall jabber-play-sound-file sound-file)))))
 
 (defun jabber-message-display (from buffer text proposed-alert)
   "Display the buffer where a new message has arrived."
@@ -356,7 +353,7 @@ Examples:
 (defun jabber-muc-wave (nick group buffer text proposed-alert)
   "Play the wave file specified in `jabber-alert-muc-wave'"
   (when proposed-alert
-    (play-sound-file jabber-alert-muc-wave)))
+    (funcall jabber-play-sound-file jabber-alert-muc-wave)))
 
 (defun jabber-muc-display (nick group buffer text proposed-alert)
   "Display the buffer where a new message has arrived."
@@ -414,7 +411,7 @@ This function is not called directly, but is the default for
 			       (return (cdr entry))))
 			   jabber-alert-presence-wave)))
       (unless (equal sound-file "")
-	(play-sound-file sound-file)))))
+	(funcall jabber-play-sound-file sound-file)))))
 
 ;; This is now defined in jabber-roster.el.
 ;; (defun jabber-presence-update-roster (who oldstatus newstatus statustext proposed-alert)
@@ -444,7 +441,7 @@ This function uses `jabber-info-message-alist' to find a message."
 (defun jabber-info-wave (infotype buffer proposed-alert)
   "Play the wave file specified in `jabber-alert-info-wave'"
   (if proposed-alert
-      (play-sound-file jabber-alert-info-wave)))
+      (funcall jabber-play-sound-file jabber-alert-info-wave)))
 
 (defun jabber-info-display (infotype buffer proposed-alert)
   "Display buffer of completed request"
@@ -456,6 +453,24 @@ This function uses `jabber-info-message-alist' to find a message."
   (when proposed-alert
     (switch-to-buffer buffer)))
 
+;;; Personal alert hooks
+(defmacro define-personal-jabber-alert (name)
+  "From ALERT function, make ALERT-personal function. Makes sence only for MUC."  
+  (let ((sn (symbol-name name)))
+    (let ((func (intern (format "%s-personal" sn))))
+    `(progn
+       (defun ,func (nick group buffer text proposed-alert)
+         (if (jabber-muc-looks-like-personal-p text group)
+             (,name nick group buffer text proposed-alert)))
+       (pushnew (quote ,func) (get 'jabber-alert-muc-hooks 'custom-options)))))
+  )
+
+(define-personal-jabber-alert jabber-muc-beep)
+(define-personal-jabber-alert jabber-muc-wave)
+(define-personal-jabber-alert jabber-muc-echo)
+(define-personal-jabber-alert jabber-muc-switch)
+(define-personal-jabber-alert jabber-muc-display)
+
 (provide 'jabber-alert)
 
 ;;; arch-tag: 725bd73e-c613-4fdc-a11d-3392a7598d4f
