@@ -1,6 +1,6 @@
 ;; jabber-chat.el - one-to-one chats
 
-;; Copyright (C) 2005 - Magnus Henoch - mange@freemail.hu
+;; Copyright (C) 2005, 2007, 2008 - Magnus Henoch - mange@freemail.hu
 
 ;; This file is a part of jabber.el.
 
@@ -21,6 +21,7 @@
 (require 'jabber-core)
 (require 'jabber-chatbuffer)
 (require 'jabber-history)
+(require 'jabber-autoloads)
 
 (require 'ewoc)
 (eval-when-compile (require 'cl))
@@ -41,9 +42,11 @@ with):
   :group 'jabber-chat)
 
 (defcustom jabber-chat-header-line-format
-  '("" (:eval (let ((buddy (jabber-jid-symbol jabber-chatting-with)))
-		(jabber-propertize " "
-				   'display (get buddy 'avatar))))
+  '("" (jabber-chat-buffer-show-avatar
+	(:eval 
+	 (let ((buddy (jabber-jid-symbol jabber-chatting-with)))
+	   (jabber-propertize " "
+			      'display (get buddy 'avatar)))))
     (:eval (jabber-jid-displayname jabber-chatting-with))
     "\t" (:eval (let ((buddy (jabber-jid-symbol jabber-chatting-with)))
 		  (propertize 
@@ -54,13 +57,21 @@ with):
 		   (or (cdr (assoc (get buddy 'show) jabber-presence-faces))
 		       'jabber-roster-user-online))))
     "\t" (:eval (jabber-fix-status (get (jabber-jid-symbol jabber-chatting-with) 'status)))
-    "\t" jabber-events-message)		;see jabber-events.el
+    "\t" jabber-events-message		;see jabber-events.el
+    "\t" jabber-chatstates-message)		;see jabber-chatstates.el
   "The specification for the header line of chat buffers.
 
 The format is that of `mode-line-format' and `header-line-format'."
   :type 'sexp
   :group 'jabber-chat)
 
+(defcustom jabber-chat-buffer-show-avatar t
+  "Show avatars in header line of chat buffer?
+This variable might not take effect if you have changed
+`jabber-chat-header-line-format'."
+  :type 'boolean
+  :group 'jabber-chat)
+
 (defcustom jabber-chat-time-format "%H:%M"
   "The format specification for instant messages in the chat buffer.
 See also `jabber-chat-delayed-time-format'.
@@ -103,8 +114,8 @@ These fields are available:
 
 %t   Time, formatted according to `jabber-chat-time-format'
      or `jabber-chat-delayed-time-format'
-%n   Nickname (`jabber-nickname')
 %u   Username
+%n   Nickname (obsolete, same as username)
 %r   Resource
 %j   Bare JID (without resource)"
   :type 'string
@@ -157,6 +168,7 @@ These fields are available:
   "Face used for error messages"
   :group 'jabber-chat)
 
+;;;###autoload
 (defvar jabber-chatting-with nil
   "JID of the person you are chatting with")
 
@@ -199,6 +211,7 @@ added to the outgoing message.")
   "Float-time of earliest backlog entry inserted into buffer.
 nil if no backlog has been inserted.")
 
+;;;###autoload
 (defun jabber-chat-get-buffer (chat-with)
   "Return the chat buffer for chatting with CHAT-WITH (bare or full JID).
 Either a string or a buffer is returned, so use `get-buffer' or
@@ -207,14 +220,16 @@ Either a string or a buffer is returned, so use `get-buffer' or
 	       (list
 		(cons ?n (jabber-jid-displayname chat-with))
 		(cons ?j (jabber-jid-user chat-with))
-		(cons ?r (jabber-jid-resource chat-with)))))
+		(cons ?r (or (jabber-jid-resource chat-with) "")))))
 
-(defun jabber-chat-create-buffer (chat-with)
+(defun jabber-chat-create-buffer (jc chat-with)
   "Prepare a buffer for chatting with CHAT-WITH.
 This function is idempotent."
   (with-current-buffer (get-buffer-create (jabber-chat-get-buffer chat-with))
     (unless (eq major-mode 'jabber-chat-mode)
-      (jabber-chat-mode #'jabber-chat-pp))
+      (jabber-chat-mode jc #'jabber-chat-pp))
+    ;; Make sure the connection variable is up to date.
+    (setq jabber-buffer-connection jc)
 
     (make-local-variable 'jabber-chatting-with)
     (setq jabber-chatting-with chat-with)
@@ -279,7 +294,7 @@ This function is idempotent."
 
 (add-to-list 'jabber-message-chain 'jabber-process-chat)
 
-(defun jabber-process-chat (xml-data)
+(defun jabber-process-chat (jc xml-data)
   "If XML-DATA is a one-to-one chat message, handle it as such."
   ;; XXX: there's more to being a chat message than not being MUC.
   ;; Maybe make independent predicate.
@@ -292,9 +307,10 @@ This function is idempotent."
 				 xml-data 'body))))))
       (with-current-buffer (if (jabber-muc-sender-p from)
 			       (jabber-muc-private-create-buffer
+				jc
 				(jabber-jid-user from)
 				(jabber-jid-resource from))
-			     (jabber-chat-create-buffer from))
+			     (jabber-chat-create-buffer jc from))
 	;; Call alert hooks only when something is output
 	(when (or error-p
 		  (run-hook-with-args-until-success 'jabber-chat-printers xml-data :foreign :printp))
@@ -308,8 +324,8 @@ This function is idempotent."
 				(funcall jabber-alert-message-function 
 					 from (current-buffer) body-text))))))))
 
-(defun jabber-chat-send (body)
-  "Send BODY, and display it in chat buffer."
+(defun jabber-chat-send (jc body)
+  "Send BODY through connection JC, and display it in chat buffer."
   ;; Build the stanza...
   (let* ((id (apply 'format "emacs-msg-%d.%d.%d" (current-time)))
 	 (stanza-to-send `(message 
@@ -325,7 +341,7 @@ This function is idempotent."
       (jabber-maybe-print-rare-time
        (ewoc-enter-last jabber-chat-ewoc (list :local stanza-to-send :time (current-time)))))
     ;; ...and send it...
-    (jabber-send-sexp stanza-to-send)))
+    (jabber-send-sexp jc stanza-to-send)))
 
 (defun jabber-chat-pp (data)
   "Pretty-print a <message/> stanza.
@@ -336,7 +352,14 @@ This function is used as an ewoc prettyprinter."
 	 (original-timestamp (when (listp (cadr data))
 			       (jabber-xml-path (cadr data) '(("jabber:x:delay" . "x")))))
 	 (internal-time
-	  (plist-get (cddr data) :time)))
+	  (plist-get (cddr data) :time))
+	 (body (ignore-errors (car
+	       (jabber-xml-node-children
+		(car
+		 (jabber-xml-get-children (cadr data) 'body))))))
+	 (/me-p
+	  (and (> (length body) 4)
+	       (string= (substring body 0 4) "/me "))))
 
     ;; Print prompt...
     (let ((delayed (or original-timestamp (plist-get (cddr data) :delayed))))
@@ -344,20 +367,22 @@ This function is used as an ewoc prettyprinter."
 	(:local
 	 (jabber-chat-self-prompt (or (jabber-x-delay original-timestamp)
 				      internal-time)
-				  delayed))
+				  delayed
+				  /me-p))
 	(:foreign
 	 ;; For :error and :notice, this might be a string... beware
 	 (jabber-chat-print-prompt (when (listp (cadr data)) (cadr data)) 
 				   (or (jabber-x-delay original-timestamp)
 				       internal-time)
-				   delayed))
-	((:error :notice)
+				   delayed
+				   /me-p))
+	((:error :notice :subscription-request)
 	 (jabber-chat-system-prompt (or (jabber-x-delay original-timestamp)
 					internal-time)))
 	(:muc-local
-	 (jabber-muc-print-prompt (cadr data) t))
+	 (jabber-muc-print-prompt (cadr data) t /me-p))
         (:muc-foreign
-         (jabber-muc-print-prompt (cadr data)))
+         (jabber-muc-print-prompt (cadr data) nil /me-p))
 	((:muc-notice :muc-error)
 	 (jabber-muc-system-prompt))))
     
@@ -376,7 +401,24 @@ This function is used as an ewoc prettyprinter."
        (insert (cadr data)))
       (:rare-time
        (insert (jabber-propertize (format-time-string jabber-rare-time-format (cadr data))
-				  'face 'jabber-rare-time-face))))
+				  'face 'jabber-rare-time-face)))
+      (:subscription-request
+       (insert "This user requests subscription to your presence.\n")
+       (when (and (stringp (cadr data)) (not (zerop (length (cadr data)))))
+	 (insert "Message: " (cadr data) "\n"))
+       (insert "Accept?\n\n")
+       (flet ((button
+	       (text action)
+	       (if (fboundp 'insert-button)
+		   (insert-button text 'action action)
+		 ;; simple button replacement
+		 (let ((keymap (make-keymap)))
+		   (define-key keymap "\r" action)
+		   (insert (jabber-propertize text 'keymap keymap 'face 'highlight))))
+	       (insert "\t")))
+	 (button "Mutual" 'jabber-subscription-accept-mutual)
+	 (button "One-way" 'jabber-subscription-accept-one-way)
+	 (button "Decline" 'jabber-subscription-decline))))
 
     (when jabber-chat-fill-long-lines
       (save-restriction
@@ -409,13 +451,14 @@ This function is used as an ewoc prettyprinter."
 					    (jabber-message-time data))))
       (ewoc-enter-before jabber-chat-ewoc node (list :rare-time (jabber-message-time data))))))
 
-(defun jabber-chat-print-prompt (xml-data timestamp delayed)
+(defun jabber-chat-print-prompt (xml-data timestamp delayed dont-print-nick-p)
   "Print prompt for received message in XML-DATA.
 TIMESTAMP is the timestamp to print, or nil to get it
 from a jabber:x:delay element.
 If DELAYED is true, print long timestamp
 \(`jabber-chat-delayed-time-format' as opposed to
-`jabber-chat-time-format')."
+`jabber-chat-time-format').
+If DONT-PRINT-NICK-P is true, don't include nickname."
   (let ((from (jabber-xml-get-attribute xml-data 'from))
 	(timestamp (or timestamp
 		       (car (delq nil (mapcar 'jabber-x-delay (jabber-xml-get-children xml-data 'x)))))))
@@ -427,7 +470,7 @@ If DELAYED is true, print long timestamp
 					 jabber-chat-delayed-time-format
 				       jabber-chat-time-format)
 				     timestamp))
-			   (cons ?n (jabber-jid-displayname from))
+			   (cons ?n (if dont-print-nick-p "" (jabber-jid-displayname from)))
 			   (cons ?u (or (jabber-jid-username from) from))
 			   (cons ?r (jabber-jid-resource from))
 			   (cons ?j (jabber-jid-user from))))
@@ -449,27 +492,33 @@ If DELAYED is true, print long timestamp
 	   'help-echo
 	   (concat (format-time-string "System message on %Y-%m-%d %H:%M:%S" timestamp)))))
 
-(defun jabber-chat-self-prompt (timestamp delayed)
+(defun jabber-chat-self-prompt (timestamp delayed dont-print-nick-p)
   "Print prompt for sent message.
 TIMESTAMP is the timestamp to print, or nil for now.
 If DELAYED is true, print long timestamp
 \(`jabber-chat-delayed-time-format' as opposed to
-`jabber-chat-time-format')."
-  (insert (jabber-propertize 
-	   (format-spec jabber-chat-local-prompt-format
-			(list
-			 (cons ?t (format-time-string 
-				   (if delayed
-				       jabber-chat-delayed-time-format
-				     jabber-chat-time-format)
-				   timestamp))
-			 (cons ?n jabber-nickname)
-			 (cons ?u jabber-username)
-			 (cons ?r jabber-resource)
-			 (cons ?j (concat jabber-username "@" jabber-server))))
-	   'face 'jabber-chat-prompt-local
-	   'help-echo
-	   (concat (format-time-string "On %Y-%m-%d %H:%M:%S" timestamp) " from you"))))
+`jabber-chat-time-format').
+If DONT-PRINT-NICK-P is true, don't include nickname."
+  (let* ((state-data (fsm-get-state-data jabber-buffer-connection))
+	 (username (plist-get state-data :username))
+	 (server (plist-get state-data :server))
+	 (resource (plist-get state-data :resource))
+	 (nickname username))
+    (insert (jabber-propertize 
+	     (format-spec jabber-chat-local-prompt-format
+			  (list
+			   (cons ?t (format-time-string 
+				     (if delayed
+					 jabber-chat-delayed-time-format
+				       jabber-chat-time-format)
+				     timestamp))
+			   (cons ?n (if dont-print-nick-p "" nickname))
+			   (cons ?u username)
+			   (cons ?r resource)
+			   (cons ?j (concat username "@" server))))
+	     'face 'jabber-chat-prompt-local
+	     'help-echo
+	     (concat (format-time-string "On %Y-%m-%d %H:%M:%S" timestamp) " from you")))))
 
 (defun jabber-chat-print-error (xml-data)
   "Print error in given <message/> in a readable way."
@@ -509,11 +558,12 @@ If DELAYED is true, print long timestamp
     (when body
 
       (when (eql mode :insert)
-	(if (string-match "^/me \\(.*\\)$" body)
-	    (let ((action (match-string 1 body))
+	(if (and (> (length body) 4)
+		 (string= (substring body 0 4) "/me "))
+	    (let ((action (substring body 4))
 		  (nick (cond
 			 ((eq who :local)
-			  jabber-nickname)
+			  (plist-get (fsm-get-state-data jabber-buffer-connection) :username))
 			 ((jabber-muc-message-p xml-data)
 			  (jabber-jid-resource (jabber-xml-get-attribute xml-data 'from)))
 			 (t
@@ -526,8 +576,8 @@ If DELAYED is true, print long timestamp
 	  (insert (jabber-propertize 
 		   body
 		   'face (case who
-			   (:foreign 'jabber-chat-text-foreign)
-			   (:local 'jabber-chat-text-local))))))
+			   ((:foreign :muc-foreign) 'jabber-chat-text-foreign)
+			   ((:local :muc-local) 'jabber-chat-text-local))))))
       t)))
 
 (defun jabber-chat-print-url (xml-data who mode)
@@ -551,19 +601,22 @@ If DELAYED is true, print long timestamp
 
 (defun jabber-chat-goto-address (&rest ignore)
   "Call `goto-address' on the newly written text."
-  (goto-address))
+  (ignore-errors 
+    (goto-address)))
 
 ;; jabber-compose is autoloaded in jabber.el
 (add-to-list 'jabber-jid-chat-menu
 	     (cons "Compose message" 'jabber-compose))
 
-(defun jabber-send-message (to subject body type)
+(defun jabber-send-message (jc to subject body type)
   "send a message tag to the server"
-  (interactive (list (jabber-read-jid-completing "to: ")
+  (interactive (list (jabber-read-account)
+		     (jabber-read-jid-completing "to: ")
 		     (jabber-read-with-input-method "subject: ")
 		     (jabber-read-with-input-method "body: ")
 		     (read-string "type: ")))
-  (jabber-send-sexp `(message ((to . ,to)
+  (jabber-send-sexp jc
+		    `(message ((to . ,to)
                                ,(if (> (length type) 0)
                                     `(type . ,type)))
                               ,(if (> (length subject) 0)
@@ -576,13 +629,17 @@ If DELAYED is true, print long timestamp
 (add-to-list 'jabber-jid-chat-menu
 	     (cons "Start chat" 'jabber-chat-with))
 
-(defun jabber-chat-with (jid &optional other-window)
+(defun jabber-chat-with (jc jid &optional other-window)
   "Open an empty chat window for chatting with JID.
 With a prefix argument, open buffer in other window.
 Returns the chat buffer."
-  (interactive (list (jabber-read-jid-completing "chat with:")
-		     current-prefix-arg))
-  (let ((buffer (jabber-chat-create-buffer jid)))
+  (interactive (let ((jid
+		      (jabber-read-jid-completing "chat with:"))
+		     (account
+		      (jabber-read-account)))
+		 (list 
+		  account jid current-prefix-arg)))
+  (let ((buffer (jabber-chat-create-buffer jc jid)))
     (if other-window
 	(switch-to-buffer-other-window buffer)
       (switch-to-buffer buffer))))
@@ -593,9 +650,11 @@ Signal an error if there is no JID at point.
 With a prefix argument, open buffer in other window."
   (interactive "P")
   (let ((jid-at-point (get-text-property (point)
-					 'jabber-jid)))
-    (if jid-at-point
-	(jabber-chat-with jid-at-point other-window)
+					 'jabber-jid))
+	(account (get-text-property (point)
+				    'jabber-account)))
+    (if (and jid-at-point account)
+	(jabber-chat-with account jid-at-point other-window)
       (error "No contact at point"))))
 
 (provide 'jabber-chat)
