@@ -1,6 +1,6 @@
 ;; jabber-chatbuffer.el - functions common to all chat buffers
 
-;; Copyright (C) 2005 - Magnus Henoch - mange@freemail.hu
+;; Copyright (C) 2005, 2007, 2008 - Magnus Henoch - mange@freemail.hu
 
 ;; This file is a part of jabber.el.
 
@@ -41,12 +41,20 @@ window or at `fill-column', whichever is shorter."
 (defvar jabber-chat-ewoc nil
   "The ewoc showing the messages of this chat buffer.")
 
-(defun jabber-chat-mode (ewoc-pp)
+;;;###autoload
+(defvar jabber-buffer-connection nil
+  "The connection used by this buffer.")
+;;;###autoload
+(make-variable-buffer-local 'jabber-buffer-connection)
+
+(defun jabber-chat-mode (jc ewoc-pp)
   "\\{jabber-chat-mode-map}"
   (kill-all-local-variables)
   ;; Make sure to set this variable somewhere
   (make-local-variable 'jabber-send-function)
 
+  (setq jabber-buffer-connection jc)
+
   (make-local-variable 'scroll-conservatively)
   (setq scroll-conservatively 5)
 
@@ -88,12 +96,21 @@ window or at `fill-column', whichever is shorter."
 
 (defun jabber-chat-buffer-send ()
   (interactive)
-  (let ((body (delete-and-extract-region jabber-point-insert (point-max))))
-    ;; If user accidentally hits RET without writing anything,
-    ;; delete-and-extract-region returns "".  In that case,
-    ;; no message should be sent.
-    (unless (zerop (length body))
-      (funcall jabber-send-function body))))
+  ;; If user accidentally hits RET without writing anything, just
+  ;; ignore it.
+  (when (plusp (- (point-max) jabber-point-insert))
+    ;; If connection was lost...
+    (unless (memq jabber-buffer-connection jabber-connections)
+      ;; ...maybe there is a new connection to the same account.
+      (let ((new-jc (jabber-find-active-connection jabber-buffer-connection)))
+	(if new-jc
+	    ;; If so, just use it.
+	    (setq jabber-buffer-connection new-jc)
+	  ;; Otherwise, ask for a new account.
+	  (setq jabber-buffer-connection (jabber-read-account t)))))
+
+    (let ((body (delete-and-extract-region jabber-point-insert (point-max))))
+      (funcall jabber-send-function jabber-buffer-connection body))))
 
 (defun jabber-chat-buffer-fill-long-lines ()
   "Fill lines that are wider than the window width."
