@@ -1,7 +1,7 @@
 ;; jabber-muc.el - advanced MUC functions
 
+;; Copyright (C) 2003, 2004, 2007, 2008 - Magnus Henoch - mange@freemail.hu
 ;; Copyright (C) 2002, 2003, 2004 - tom berger - object@intelectronica.net
-;; Copyright (C) 2003, 2004 - Magnus Henoch - mange@freemail.hu
 
 ;; This file is a part of jabber.el.
 
@@ -22,9 +22,11 @@
 (require 'jabber-chat)
 (require 'jabber-widget)
 (require 'jabber-newdisco)
+(require 'jabber-autoloads)
 
 (require 'cl)
 
+;;;###autoload
 (defvar *jabber-active-groupchats* nil
   "alist of groupchats and nicknames
 Keys are strings, the bare JID of the room.
@@ -47,6 +49,7 @@ Values are lists of nickname strings.")
 (defvar jabber-muc-topic ""
   "The topic of the current MUC room.")
 
+;;;###autoload
 (defcustom jabber-muc-default-nicknames nil
   "Default nickname for specific MUC rooms."
   :group 'jabber-chat
@@ -55,8 +58,12 @@ Values are lists of nickname strings.")
 		(string :tag "JID of room")
 		(string :tag "Nickname"))))
 
+;;;###autoload
 (defcustom jabber-muc-autojoin nil
-  "List of MUC rooms to automatically join on connection."
+  "List of MUC rooms to automatically join on connection.
+This list is saved in your Emacs customizations.  You can also store
+such a list on the Jabber server, where it is available to every
+client; see `jabber-edit-bookmarks'."
   :group 'jabber-chat
   :type '(repeat (string :tag "JID of room")))
 
@@ -124,17 +131,20 @@ These fields are available:
 (defcustom jabber-muc-private-header-line-format
   '(" " (:eval (jabber-jid-resource jabber-chatting-with))
     " in " (:eval (jabber-jid-displayname (jabber-jid-user jabber-chatting-with)))
-    "\t" jabber-events-message)
+    "\t" jabber-events-message
+    "\t" jabber-chatstates-message)
   "The specification for the header line of private MUC chat buffers.
 
 The format is that of `mode-line-format' and `header-line-format'."
   :type 'sexp
   :group 'jabber-chat)
 
+;;;###autoload
 (defvar jabber-muc-printers '()
   "List of functions that may be able to print part of a MUC message.
 This gets prepended to `jabber-chat-printers', which see.")
 
+;;;###autoload
 (defun jabber-muc-get-buffer (group)
   "Return the chat buffer for chatroom GROUP.
 Either a string or a buffer is returned, so use `get-buffer' or
@@ -144,12 +154,14 @@ Either a string or a buffer is returned, so use `get-buffer' or
 		(cons ?n (jabber-jid-displayname group))
 		(cons ?j (jabber-jid-user group)))))
 
-(defun jabber-muc-create-buffer (group)
+(defun jabber-muc-create-buffer (jc group)
   "Prepare a buffer for chatroom GROUP.
 This function is idempotent."
   (with-current-buffer (get-buffer-create (jabber-muc-get-buffer group))
     (unless (eq major-mode 'jabber-chat-mode)
-      (jabber-chat-mode #'jabber-chat-pp))
+      (jabber-chat-mode jc #'jabber-chat-pp))
+    ;; Make sure the connection variable is up to date.
+    (setq jabber-buffer-connection jc)
 
     (set (make-local-variable 'jabber-group) group)
     (make-local-variable 'jabber-muc-topic)
@@ -157,6 +169,7 @@ This function is idempotent."
     (setq header-line-format jabber-muc-header-line-format)
     (current-buffer)))
 
+;;;###autoload
 (defun jabber-muc-private-get-buffer (group nickname)
   "Return the chat buffer for private chat with NICKNAME in GROUP.
 Either a string or a buffer is returned, so use `get-buffer' or
@@ -166,12 +179,12 @@ Either a string or a buffer is returned, so use `get-buffer' or
 		(cons ?g (jabber-jid-displayname group))
 		(cons ?n nickname))))
 
-(defun jabber-muc-private-create-buffer (group nickname)
+(defun jabber-muc-private-create-buffer (jc group nickname)
   "Prepare a buffer for chatting with NICKNAME in GROUP.
 This function is idempotent."
   (with-current-buffer (get-buffer-create (jabber-muc-private-get-buffer group nickname))
     (unless (eq major-mode 'jabber-chat-mode)
-      (jabber-chat-mode #'jabber-chat-pp))
+      (jabber-chat-mode jc #'jabber-chat-pp))
 
     (set (make-local-variable 'jabber-chatting-with) (concat group "/" nickname))
     (setq jabber-send-function 'jabber-chat-send)
@@ -179,11 +192,12 @@ This function is idempotent."
 
     (current-buffer)))
 
-(defun jabber-muc-send (body)
+(defun jabber-muc-send (jc body)
   "Send BODY to MUC room in current buffer."
   ;; There is no need to display the sent message in the buffer, as
   ;; we will get it back from the MUC server.
-  (jabber-send-sexp `(message
+  (jabber-send-sexp jc
+		    `(message
 		      ((to . ,jabber-group)
 		       (type . "groupchat"))
 		      (body () ,body))))
@@ -234,6 +248,12 @@ are the corresponding presence fields.
 This function is only concerned with presence stanzas resulting
 in the user entering/staying in the room."
   ;; The keys in the plist are affiliation, role and jid.
+  (when (plist-get new-plist 'jid)
+    ;; nickname is only used for displaying, so we can modify it if we
+    ;; want to.
+    (setq nickname (concat nickname " <" 
+			   (jabber-jid-user (plist-get new-plist 'jid))
+			   ">")))
   (cond
    ((null old-plist)
     ;; User enters the room
@@ -303,6 +323,14 @@ in the user entering/staying in the room."
       (let ((participant (assoc nickname (cdr participants))))
 	(setf (cdr participants) (delq participant (cdr participants)))))))
 
+(defmacro jabber-muc-argument-list (&optional args)
+  "Prepend connection and group name to ARGS.
+If the current buffer is not an MUC buffer, signal an error.
+This macro is meant for use as an argument to `interactive'."
+  `(if (null jabber-group)
+       (error "Not in MUC buffer")
+     (nconc (list jabber-buffer-connection jabber-group) ,args)))
+
 (defun jabber-muc-read-completing (prompt &optional allow-not-joined)
   "Read the name of a joined chatroom, or use chatroom of current buffer, if any.
 If ALLOW-NOT-JOINED is provided and true, permit choosing any
@@ -325,16 +353,16 @@ JID; only provide completion as a guide."
 
 (add-to-list 'jabber-jid-muc-menu
    (cons "Configure groupchat" 'jabber-groupchat-get-config))
-(defun jabber-groupchat-get-config (group)
+(defun jabber-groupchat-get-config (jc group)
   "Ask for MUC configuration form"
-  (interactive (list (jabber-muc-read-completing "Configure group: " t)))
-  (jabber-send-iq group
+  (interactive (jabber-muc-argument-list))
+  (jabber-send-iq jc group
 		  "get"
 		  '(query ((xmlns . "http://jabber.org/protocol/muc#owner")))
 		  #'jabber-process-data #'jabber-groupchat-render-config
 		  #'jabber-process-data "MUC configuration request failed"))
 
-(defun jabber-groupchat-render-config (xml-data)
+(defun jabber-groupchat-render-config (jc xml-data)
   "Render MUC configuration form"
 
   (let ((query (jabber-iq-query xml-data))
@@ -346,6 +374,7 @@ JID; only provide completion as a guide."
 	(insert "No configuration possible.\n")
       
     (jabber-init-widget-buffer (jabber-xml-get-attribute xml-data 'from))
+    (setq jabber-buffer-connection jc)
 
     (jabber-render-xdata-form xdata)
 
@@ -360,7 +389,7 @@ JID; only provide completion as a guide."
 (defun jabber-groupchat-submit-config (&rest ignore)
   "Submit MUC configuration form."
 
-  (jabber-send-iq jabber-submit-to
+  (jabber-send-iq jabber-buffer-connection jabber-submit-to
 		  "set"
 		  `(query ((xmlns . "http://jabber.org/protocol/muc#owner"))
 			  ,(jabber-parse-xdata-form))
@@ -370,7 +399,7 @@ JID; only provide completion as a guide."
 (defun jabber-groupchat-cancel-config (&rest ignore)
   "Cancel MUC configuration form."
 
-  (jabber-send-iq jabber-submit-to
+  (jabber-send-iq jabber-buffer-connection jabber-submit-to
 		  "set"
 		  '(query ((xmlns . "http://jabber.org/protocol/muc#owner"))
 			  (x ((xmlns . "jabber:x:data") (type . "cancel"))))
@@ -379,25 +408,26 @@ JID; only provide completion as a guide."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Join groupchat" 'jabber-groupchat-join))
 
-(defun jabber-groupchat-join (group nickname &optional popup)
+(defun jabber-groupchat-join (jc group nickname &optional popup)
   "join a groupchat, or change nick.
 In interactive calls, or if POPUP is true, switch to the
 groupchat buffer."
   (interactive 
-   (let ((group (jabber-read-jid-completing "group: ")))
-     (list group (jabber-muc-read-my-nickname group) t)))
+   (let ((account (jabber-read-account))
+	 (group (jabber-read-jid-completing "group: ")))
+     (list account group (jabber-muc-read-my-nickname account group) t)))
 
   ;; If the user is already in the room, we don't need as many checks.
   (if (or (assoc group *jabber-active-groupchats*)
 	  ;; Or if the users asked us not to check disco info.
 	  jabber-muc-disable-disco-check)
-      (jabber-groupchat-join-3 group nickname nil popup)
+      (jabber-groupchat-join-3 jc group nickname nil popup)
     ;; Else, send a disco request to find out what we are connecting
     ;; to.
-    (jabber-disco-get-info group nil #'jabber-groupchat-join-2
+    (jabber-disco-get-info jc group nil #'jabber-groupchat-join-2
 			   (list group nickname popup))))
 
-(defun jabber-groupchat-join-2 (closure result)
+(defun jabber-groupchat-join-2 (jc closure result)
   (destructuring-bind (group nickname popup) closure
     (let ( ;; Either success...
 	  (identities (car result))
@@ -427,22 +457,25 @@ groupchat buffer."
       (let ((password
 	     ;; Is the room password-protected?
 	     (when (member "muc_passwordprotected" features)
-	       (read-passwd (format "Password for %s: " (jabber-jid-displayname group))))))
+	       (or
+		(jabber-get-conference-data jc group nil :password)
+		(read-passwd (format "Password for %s: " (jabber-jid-displayname group)))))))
 
-	(jabber-groupchat-join-3 group nickname password popup)))))
+	(jabber-groupchat-join-3 jc group nickname password popup)))))
 
-(defun jabber-groupchat-join-3 (group nickname password popup)
+(defun jabber-groupchat-join-3 (jc group nickname password popup)
 
   ;; Remember that this is a groupchat _before_ sending the stanza.
   ;; The response might come quicker than you think.
 
   (puthash (jabber-jid-symbol group) nickname jabber-pending-groupchats)
   
-  (jabber-send-sexp `(presence ((to . ,(format "%s/%s" group nickname)))
+  (jabber-send-sexp jc
+		    `(presence ((to . ,(format "%s/%s" group nickname)))
 			       (x ((xmlns . "http://jabber.org/protocol/muc"))
 				  ,@(when password
 				      `((password () ,password))))
-			       ,@(jabber-presence-children)))
+			       ,@(jabber-presence-children jc)))
 
   ;; There, stanza sent.  Now we just wait for the MUC service to
   ;; mirror the stanza.  This is handled in
@@ -452,14 +485,15 @@ groupchat buffer."
   ;; But if the user interactively asked to join, he/she probably
   ;; wants the buffer to pop up right now.
   (when popup
-    (let ((buffer (jabber-muc-create-buffer group)))
+    (let ((buffer (jabber-muc-create-buffer jc group)))
       (switch-to-buffer buffer))))
 
-(defun jabber-muc-read-my-nickname (group)
+(defun jabber-muc-read-my-nickname (jc group)
   "Read nickname for joining GROUP."
   (let ((default-nickname (or
+			   (jabber-get-conference-data jc group nil :nick)
 			   (cdr (assoc group jabber-muc-default-nicknames))
-			   jabber-nickname)))
+			   (plist-get (fsm-get-state-data jc) :username))))
     (jabber-read-with-input-method (format "Nickname: (default %s) "
 					   default-nickname) 
 				   nil nil default-nickname)))
@@ -472,30 +506,25 @@ groupchat buffer."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Leave groupchat" 'jabber-groupchat-leave))
 
-(defun jabber-groupchat-leave (group)
+(defun jabber-groupchat-leave (jc group)
   "leave a groupchat"
-  (interactive (list (jabber-muc-read-completing "Leave which group: ")))
+  (interactive (jabber-muc-argument-list))
   (let ((whichgroup (assoc group *jabber-active-groupchats*)))
     ;; send unavailable presence to our own nick in room
-    (jabber-send-sexp `(presence ((to . ,(format "%s/%s" group (cdr whichgroup)))
+    (jabber-send-sexp jc
+		      `(presence ((to . ,(format "%s/%s" group (cdr whichgroup)))
 				  (type . "unavailable"))))))
 
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "List participants" 'jabber-muc-names))
 
-(defun jabber-muc-names (group)
-  "Print names, affiliations, and roles of participants in GROUP."
-  (interactive (list (jabber-muc-read-completing "Group: ")))
-  (with-current-buffer (jabber-muc-create-buffer group)
-    (ewoc-enter-last jabber-chat-ewoc (list :notice
-					    (jabber-muc-print-names
-					     (cdr (assoc group jabber-muc-participants)))
-					    :time (current-time)))
-    ;; (let ((jabber-chat-fill-long-lines nil))
-;;       (jabber-chat-buffer-display 'jabber-muc-system-prompt nil
-;; 				  '(jabber-muc-print-names)
-;; 				  (cdr (assoc group jabber-muc-participants))))
-    ))
+(defun jabber-muc-names ()
+  "Print names, affiliations, and roles of participants in current buffer."
+  (interactive)
+  (ewoc-enter-last jabber-chat-ewoc (list :notice
+					  (jabber-muc-print-names
+					   (cdr (assoc jabber-group jabber-muc-participants)))
+					  :time (current-time))))
 
 (defun jabber-muc-print-names (participants)
   "Format and return data in PARTICIPANTS."
@@ -513,13 +542,12 @@ groupchat buffer."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Set topic" 'jabber-muc-set-topic))
 
-(defun jabber-muc-set-topic (group topic)
+(defun jabber-muc-set-topic (jc group topic)
   "Set topic of GROUP to TOPIC."
   (interactive
-   (let ((group (jabber-muc-read-completing "Group: ")))
-     (list group
-	   (jabber-read-with-input-method "New topic: " jabber-muc-topic))))
-  (jabber-send-message group topic nil "groupchat"))
+   (jabber-muc-argument-list
+    (list (jabber-read-with-input-method "New topic: " jabber-muc-topic))))
+  (jabber-send-message jc group topic nil "groupchat"))
 
 (defun jabber-muc-snarf-topic (xml-data)
   "Record subject (topic) of the given <message/>, if any."
@@ -530,16 +558,16 @@ groupchat buffer."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Set role (kick, voice, op)" 'jabber-muc-set-role))
 
-(defun jabber-muc-set-role (group nickname role reason)
+(defun jabber-muc-set-role (jc group nickname role reason)
   "Set role of NICKNAME in GROUP to ROLE, specifying REASON."
   (interactive
-   (let* ((group (jabber-muc-read-completing "Group: "))
-	  (nickname (jabber-muc-read-nickname group "Nickname: ")))
-     (list group nickname
-	   (completing-read "New role: " '(("none") ("visitor") ("participant") ("moderator")) nil t)
-	   (read-string "Reason: "))))
+   (jabber-muc-argument-list
+    (let ((nickname (jabber-muc-read-nickname jabber-group "Nickname: ")))
+      (list nickname
+	    (completing-read "New role: " '(("none") ("visitor") ("participant") ("moderator")) nil t)
+	    (read-string "Reason: ")))))
   (unless (or (zerop (length nickname)) (zerop (length role)))
-    (jabber-send-iq group "set"
+    (jabber-send-iq jc group "set"
 		    `(query ((xmlns . "http://jabber.org/protocol/muc#admin"))
 			    (item ((nick . ,nickname)
 				   (role . ,role))
@@ -551,22 +579,21 @@ groupchat buffer."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Set affiliation (ban, member, admin)" 'jabber-muc-set-affiliation))
 
-(defun jabber-muc-set-affiliation (group nickname-or-jid nickname-p affiliation reason)
+(defun jabber-muc-set-affiliation (jc group nickname-or-jid nickname-p affiliation reason)
   "Set affiliation of NICKNAME-OR-JID in GROUP to AFFILIATION.
 If NICKNAME-P is non-nil, NICKNAME-OR-JID is a nickname in the
 group, else it is a JID."
   (interactive
-   (let ((group (jabber-muc-read-completing "Group: "))
-	 (nickname-p (y-or-n-p "Specify user by room nickname? ")))
-     (list
-      group
-      (if nickname-p
-	  (jabber-muc-read-nickname group "Nickname: ")
-	(jabber-read-jid-completing "User: "))
-      nickname-p
-      (completing-read "New affiliation: "
-		       '(("none") ("outcast") ("member") ("admin") ("owner")) nil t)
-      (read-string "Reason: "))))
+   (jabber-muc-argument-list
+    (let ((nickname-p (y-or-n-p "Specify user by room nickname? ")))
+      (list
+       (if nickname-p
+	   (jabber-muc-read-nickname jabber-group "Nickname: ")
+	 (jabber-read-jid-completing "User: "))
+       nickname-p
+       (completing-read "New affiliation: "
+			'(("none") ("outcast") ("member") ("admin") ("owner")) nil t)
+       (read-string "Reason: ")))))
   (let ((jid
 	 (if nickname-p
 	     (let ((participants (cdr (assoc group jabber-muc-participants))))
@@ -578,7 +605,7 @@ group, else it is a JID."
 		 (or (plist-get participant 'jid)
 		     (error "JID of %s in group %s is unknown" nickname-or-jid group))))
 	   nickname-or-jid)))
-    (jabber-send-iq group "set"
+    (jabber-send-iq jc group "set"
 		    `(query ((xmlns . "http://jabber.org/protocol/muc#admin"))
 			    (item ((jid . ,jid)
 				   (affiliation . ,affiliation))
@@ -590,13 +617,15 @@ group, else it is a JID."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Invite someone to chatroom" 'jabber-muc-invite))
 
-(defun jabber-muc-invite (jid group reason)
+(defun jabber-muc-invite (jc jid group reason)
   "Invite JID to GROUP, stating REASON."
   (interactive
-   (list (jabber-read-jid-completing "Invite whom: ")
+   (list (jabber-read-account)
+	 (jabber-read-jid-completing "Invite whom: ")
 	 (jabber-muc-read-completing "To group: ")
 	 (jabber-read-with-input-method "Reason: ")))
   (jabber-send-sexp
+   jc
    `(message ((to . ,group))
 	     (x ((xmlns . "http://jabber.org/protocol/muc#user"))
 		(invite ((to . ,jid))
@@ -626,8 +655,8 @@ group, else it is a JID."
 
 	      (let ((action
 		     `(lambda (&rest ignore) (interactive)
-			(jabber-groupchat-join ,group
-					       (jabber-muc-read-my-nickname ,group)))))
+			(jabber-groupchat-join jabber-buffer-connection ,group
+					       (jabber-muc-read-my-nickname ,jabber-buffer-connection ,group)))))
 		(if (fboundp 'insert-button)
 		    (insert-button "Accept"
 				   'action action)
@@ -646,6 +675,7 @@ group, else it is a JID."
 				 (jabber-read-with-input-method
 				  "Reason: ")))
 			    (jabber-send-sexp
+			     jabber-buffer-connection
 			     (list 'message
 				   (list (cons 'to ,group))
 				   (list 'x
@@ -665,14 +695,26 @@ group, else it is a JID."
 						 'face 'highlight))))))))
 	  (return t))))))
 
-(defun jabber-muc-autojoin ()
-  "Join rooms specified in variable `jabber-muc-autojoin'."
-  (interactive)
-  (dolist (group jabber-muc-autojoin)
-    (jabber-groupchat-join group (or
-				  (cdr (assoc group jabber-muc-default-nicknames))
-				  jabber-nickname))))
+(defun jabber-muc-autojoin (jc)
+  "Join rooms specified in account bookmarks and global `jabber-muc-autojoin'."
+  (interactive (list (jabber-read-account)))
+  (let ((nickname (plist-get (fsm-get-state-data jc) :username)))
+    (when (bound-and-true-p jabber-muc-autojoin)
+      (dolist (group jabber-muc-autojoin)
+	(jabber-groupchat-join jc group (or
+					 (cdr (assoc group jabber-muc-default-nicknames))
+					 (plist-get (fsm-get-state-data jc) :username)))))
+    (jabber-get-bookmarks
+     jc
+     (lambda (jc bookmarks)
+       (dolist (bookmark bookmarks)
+	 (setq bookmark (jabber-parse-conference-bookmark bookmark))
+	 (when (and bookmark (plist-get bookmark :autojoin))
+	   (jabber-groupchat-join jc (plist-get bookmark :jid)
+				  (or (plist-get bookmark :nick)
+				      (plist-get (fsm-get-state-data jc) :username)))))))))
 
+;;;###autoload
 (defun jabber-muc-message-p (message)
   "Return non-nil if MESSAGE is a groupchat message.
 That does not include private messages in a groupchat, but does
@@ -688,11 +730,13 @@ include groupchat invites."
 	  (gethash (jabber-jid-symbol from) jabber-pending-groupchats))
      (jabber-xml-path message '(("http://jabber.org/protocol/muc#user" . "x") invite)))))
 
+;;;###autoload
 (defun jabber-muc-sender-p (jid)
   "Return non-nil if JID is a full JID of an MUC participant."
   (and (assoc (jabber-jid-user jid) *jabber-active-groupchats*)
        (jabber-jid-resource jid)))
 
+;;;###autoload
 (defun jabber-muc-private-message-p (message)
   "Return non-nil if MESSAGE is a private message in a groupchat."
   (let ((from (jabber-xml-get-attribute message 'from))
@@ -704,13 +748,12 @@ include groupchat invites."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Open private chat" 'jabber-muc-private))
 
-(defun jabber-muc-private (group nickname)
+(defun jabber-muc-private (jc group nickname)
   "Open private chat with NICKNAME in GROUP."
   (interactive
-   (let* ((group (jabber-muc-read-completing "Group: "))
-	  (nickname (jabber-muc-read-nickname group "Nickname: ")))
-     (list group nickname)))
-  (switch-to-buffer (jabber-muc-private-create-buffer group nickname)))
+   (jabber-muc-argument-list
+    (list (jabber-muc-read-nickname jabber-group "Nickname: "))))
+  (switch-to-buffer (jabber-muc-private-create-buffer jabber-buffer-connection group nickname)))
 
 (defun jabber-muc-presence-p (presence)
   "Return non-nil if PRESENCE is presence from groupchat."
@@ -734,7 +777,7 @@ Return nil if X-MUC is nil."
 			(jabber-xml-node-attributes
 			 (car (jabber-xml-get-children x-muc 'item))))))
 
-(defun jabber-muc-print-prompt (xml-data &optional local)
+(defun jabber-muc-print-prompt (xml-data &optional local dont-print-nick-p)
   "Print MUC prompt for message in XML-DATA."
   (let ((nick (jabber-jid-resource (jabber-xml-get-attribute xml-data 'from)))
 	(timestamp (car (delq nil (mapcar 'jabber-x-delay (jabber-xml-get-children xml-data 'x))))))
@@ -747,7 +790,7 @@ Return nil if X-MUC is nil."
 					     jabber-chat-delayed-time-format
 					   jabber-chat-time-format)
 					 timestamp))
-			       (cons ?n nick)
+			       (cons ?n (if dont-print-nick-p "" nick))
 			       (cons ?u nick)
 			       (cons ?r nick)
 			       (cons ?j (concat jabber-group "/" nick))))
@@ -790,7 +833,7 @@ Return nil if X-MUC is nil."
 
 (add-to-list 'jabber-message-chain 'jabber-muc-process-message)
 
-(defun jabber-muc-process-message (xml-data)
+(defun jabber-muc-process-message (jc xml-data)
   "If XML-DATA is a groupchat message, handle it as such."
   (when (jabber-muc-message-p xml-data)
     (let* ((from (jabber-xml-get-attribute xml-data 'from))
@@ -808,21 +851,27 @@ Return nil if X-MUC is nil."
 
 	   (printers (append jabber-muc-printers jabber-chat-printers)))
 
-      (with-current-buffer (jabber-muc-create-buffer group)
+      (with-current-buffer (jabber-muc-create-buffer jc group)
 	(jabber-muc-snarf-topic xml-data)
 	;; Call alert hooks only when something is output
 	(when (or error-p
 		  (run-hook-with-args-until-success 'printers xml-data type :printp))
 	  (jabber-maybe-print-rare-time
 	   (ewoc-enter-last jabber-chat-ewoc (list type xml-data :time (current-time))))
-	  
-	  (dolist (hook '(jabber-muc-hooks jabber-alert-muc-hooks))
-	    (run-hook-with-args hook
-				nick group (current-buffer) body-text
-				(funcall jabber-alert-muc-function
-					 nick group (current-buffer) body-text))))))))
 
-(defun jabber-muc-process-presence (presence)
+	  ;; ...except if the message is part of history, in which
+	  ;; case we don't want an alert.
+	  (let ((children-namespaces (mapcar (lambda (x) (when (listp x) (jabber-xml-get-attribute x 'xmlns)))
+					     (jabber-xml-node-children xml-data))))
+	    (unless (or (member "urn:xmpp:delay" children-namespaces)
+			(member "jabber:x:delay" children-namespaces))
+	      (dolist (hook '(jabber-muc-hooks jabber-alert-muc-hooks))
+		(run-hook-with-args hook
+				    nick group (current-buffer) body-text
+				    (funcall jabber-alert-muc-function
+					     nick group (current-buffer) body-text))))))))))
+
+(defun jabber-muc-process-presence (jc presence)
   (let* ((from (jabber-xml-get-attribute presence 'from))
 	 (type (jabber-xml-get-attribute presence 'type))
 	 (x-muc (find-if 
@@ -835,31 +884,45 @@ Return nil if X-MUC is nil."
 	 (item (car (jabber-xml-get-children x-muc 'item)))
 	 (actor (jabber-xml-get-attribute (car (jabber-xml-get-children item 'actor)) 'jid))
 	 (reason (car (jabber-xml-node-children (car (jabber-xml-get-children item 'reason)))))
-	 (status-code (jabber-xml-get-attribute
-		       (car (jabber-xml-get-children x-muc 'status))
-		       'code))
-	 (error-node (car (jabber-xml-get-children presence 'error))))
+	 (error-node (car (jabber-xml-get-children presence 'error)))
+	 (status-code (if error-node
+			  (jabber-xml-get-attribute error-node 'code)
+			(jabber-xml-get-attribute (car (jabber-xml-get-children x-muc 'status)) 'code))))
     ;; handle leaving a room
     (cond 
      ((or (string= type "unavailable") (string= type "error"))
-      ;; are we leaving?
-      (if (string= nickname (gethash (jabber-jid-symbol group) jabber-pending-groupchats))
-	  (let ((message (cond
-			  ((string= type "error")
-			   (concat "Error entering room"
-				   (when error-node
-				     (concat ": " (jabber-parse-error error-node)))))
-			  ((equal status-code "301")
-			   (concat "You have been banned"
-				   (when actor (concat " by " actor))
-				   (when reason (concat " - '" reason "'"))))
-			  ((equal status-code "307")
-			   (concat "You have been kicked"
-				   (when actor (concat " by " actor))
-				   (when reason (concat " - '" reason "'"))))
-			  (t
-			   "You have left the chatroom"))))
-	    (jabber-muc-remove-groupchat group)
+      ;; error from room itself? or are we leaving?
+      (if (or (null nickname)
+	      (string= nickname (gethash (jabber-jid-symbol group) jabber-pending-groupchats)))
+	  ;; Assume that an error means that we were thrown out of the
+	  ;; room...
+	  (let* ((leavingp t)
+		 (message (cond
+			   ((string= type "error")
+			    (cond
+			     ;; ...except for certain cases.
+			     ((or (equal status-code "406")
+				  (equal status-code "409"))
+			      (setq leavingp nil)
+			      (concat "Nickname change not allowed"
+				      (when error-node
+					(concat ": " (jabber-parse-error error-node)))))
+			     (t
+			      (concat "Error entering room"
+				      (when error-node
+					(concat ": " (jabber-parse-error error-node)))))))
+			   ((equal status-code "301")
+			    (concat "You have been banned"
+				    (when actor (concat " by " actor))
+				    (when reason (concat " - '" reason "'"))))
+			   ((equal status-code "307")
+			    (concat "You have been kicked"
+				    (when actor (concat " by " actor))
+				    (when reason (concat " - '" reason "'"))))
+			   (t
+			    "You have left the chatroom"))))
+	    (when leavingp
+	      (jabber-muc-remove-groupchat group))
 	    ;; If there is no buffer for this groupchat, don't bother
 	    ;; creating one just to tell that user left the room.
 	    (let ((buffer (get-buffer (jabber-muc-get-buffer group))))
@@ -874,27 +937,34 @@ Return nil if X-MUC is nil."
 					    :time (current-time)))))
 		(message "%s: %s" (jabber-jid-displayname group) message))))
 	;; or someone else?
-	(jabber-muc-remove-participant group nickname)
-	(with-current-buffer (jabber-muc-create-buffer group)
-	  (jabber-maybe-print-rare-time
-	   (ewoc-enter-last
-	    jabber-chat-ewoc
-	    (list :muc-notice
-		  (cond
-		   ((equal status-code "301")
-		    (concat nickname " has been banned"
-			    (when actor (concat " by " actor))
-			    (when reason (concat " - '" reason "'"))))
-		   ((equal status-code "307")
-		    (concat nickname " has been kicked"
-			    (when actor (concat " by " actor))
-			    (when reason (concat " - '" reason "'"))))
-		   ((equal status-code "303")
-		    (concat nickname " changes nickname to "
-			    (jabber-xml-get-attribute item 'nick)))
-		   (t
-		    (concat nickname " has left the chatroom")))
-		  :time (current-time)))))))
+	(let* ((plist (jabber-muc-participant-plist group nickname))
+	       (jid (plist-get plist 'jid))
+	       (name (concat nickname
+			     (when jid
+			       (concat " <" 
+				       (jabber-jid-user jid)
+				       ">")))))
+	  (jabber-muc-remove-participant group nickname)
+	  (with-current-buffer (jabber-muc-create-buffer jc group)
+	    (jabber-maybe-print-rare-time
+	     (ewoc-enter-last
+	      jabber-chat-ewoc
+	      (list :muc-notice
+		    (cond
+		     ((equal status-code "301")
+		      (concat name " has been banned"
+			      (when actor (concat " by " actor))
+			      (when reason (concat " - '" reason "'"))))
+		     ((equal status-code "307")
+		      (concat name " has been kicked"
+			      (when actor (concat " by " actor))
+			      (when reason (concat " - '" reason "'"))))
+		     ((equal status-code "303")
+		      (concat name " changes nickname to "
+			      (jabber-xml-get-attribute item 'nick)))
+		     (t
+		      (concat name " has left the chatroom")))
+		    :time (current-time))))))))
      (t 
       ;; someone is entering
 
@@ -915,7 +985,7 @@ Return nil if X-MUC is nil."
 	(let ((report (jabber-muc-report-delta nickname old-plist new-plist
 					       reason actor)))
 	  (when report
-	    (with-current-buffer (jabber-muc-create-buffer group)
+	    (with-current-buffer (jabber-muc-create-buffer jc group)
 	      (jabber-maybe-print-rare-time
 	       (ewoc-enter-last
 		jabber-chat-ewoc
