@@ -1,7 +1,8 @@
 ;; jabber-util.el - various utility functions    -*- coding: utf-8; -*-
 
+;; Copyright (C) 2003, 2004, 2007, 2008 - Magnus Henoch - mange@freemail.hu
 ;; Copyright (C) 2002, 2003, 2004 - tom berger - object@intelectronica.net
-;; Copyright (C) 2003, 2004 - Magnus Henoch - mange@freemail.hu
+;; Copyright (C) 2008 - Terechkov Evgenii - evg@altlinux.org
 
 ;; This file is a part of jabber.el.
 
@@ -20,17 +21,32 @@
 ;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 (eval-when-compile (require 'cl))
+(condition-case nil
+    (require 'password)
+  (error nil))
 
 (defvar jabber-jid-history nil
   "History of entered JIDs")
 
+;; Define `jabber-replace-in-string' somehow.
 (cond
- ((fboundp 'replace-in-string)
-  (defsubst jabber-replace-in-string (str regexp newtext)
-    (replace-in-string str regexp newtext t)))
+ ;; Emacs 21 has replace-regexp-in-string.
  ((fboundp 'replace-regexp-in-string)
   (defsubst jabber-replace-in-string (str regexp newtext)
-    (replace-regexp-in-string regexp newtext str t t))))
+    (replace-regexp-in-string regexp newtext str t t)))
+ ;; XEmacs has replace-in-string.  However, color-theme defines it as
+ ;; well on Emacs 2x, so this check must be last.
+ ((fboundp 'replace-in-string)
+  ;; And the version in color-theme takes only three arguments.  Check
+  ;; just to be sure.
+  (condition-case nil
+      (replace-in-string "foobar" "foo" "bar" t)
+    (wrong-number-of-arguments
+     (error "`replace-in-string' doesn't accept fourth argument")))
+  (defsubst jabber-replace-in-string (str regexp newtext)
+    (replace-in-string str regexp newtext t)))
+ (t
+  (error "No implementation of `jabber-replace-in-string' available")))
 
 ;;; XEmacs compatibility.  Stolen from ibuffer.el
 (if (fboundp 'propertize)
@@ -91,6 +107,35 @@ properties to add to the result."
  (t
   (error "No `cancel-timer' function found")))
 
+(defun jabber-concat-rosters ()
+  "Concatenate the rosters of all connected accounts."
+  (apply #'append
+	 (mapcar
+	  (lambda (jc)
+	    (plist-get (fsm-get-state-data jc) :roster))
+	  jabber-connections)))
+
+(defun jabber-connection-jid (jc)
+  "Return the full JID of the given connection."
+  (let ((sd (fsm-get-state-data jc)))
+    (concat (plist-get sd :username) "@"
+	    (plist-get sd :server) "/"
+	    (plist-get sd :resource))))
+
+(defun jabber-connection-bare-jid (jc)
+  "Return the bare JID of the given connection."
+  (let ((sd (fsm-get-state-data jc)))
+    (concat (plist-get sd :username) "@"
+	    (plist-get sd :server))))
+
+(defun jabber-find-active-connection (dead-jc)
+  "Given a dead connection, find an active connection to the same account.
+Return nil if none found."
+  (let ((jid (jabber-connection-bare-jid dead-jc)))
+    (dolist (jc jabber-connections)
+      (when (string= jid (jabber-connection-bare-jid jc))
+	(return jc)))))
+
 (defun jabber-jid-username (string)
   "return the username portion of a JID, or nil if no username"
   (when (string-match "\\(.*\\)@.*\\(/.*\\)?" string)
@@ -134,10 +179,9 @@ properties to add to the result."
     ;; XXX: "downcase" is poor man's nodeprep.  See XMPP CORE.
     (intern (downcase (jabber-jid-user string)) jabber-jid-obarray)))
 
-(defun jabber-my-jid-p (jid)
-  "Return non-nil if the specified JID is equal to the user's JID, modulo resource."
-  (equal (jabber-jid-user jid)
-	 (concat jabber-username "@" jabber-server)))
+(defun jabber-my-jid-p (jc jid)
+  "Return non-nil if the specified JID is in jabber-account-list (modulo resource). JC arg placed for compatibility and may be removed in future."
+  (member (jabber-jid-user jid) (mapcar (lambda (x) (jabber-jid-user (car x))) jabber-account-list)))
 
 (defun jabber-read-jid-completing (prompt &optional subset require-match default resource)
   "read a jid out of the current roster from the minibuffer.
@@ -163,9 +207,9 @@ bare-or-muc Turn full JIDs to bare ones, except for in MUC"
 	(completion-ignore-case t)
 	(jid-completion-table (mapcar #'(lambda (item)
 					  (cons (symbol-name item) item))
-				      (or subset *jabber-roster*)))
+				      (or subset (jabber-concat-rosters))))
 	chosen)
-    (dolist (item (or subset *jabber-roster*))
+    (dolist (item (or subset (jabber-concat-rosters)))
       (if (get item 'name)
 	  (push (cons (get item 'name) item) jid-completion-table)))
     ;; if the default is not in the allowed subset, it's not a good default
@@ -212,15 +256,71 @@ bare-or-muc Turn full JIDs to bare ones, except for in MUC"
 			     (format "(default %s) " node-at-point)))
 		 node-at-point)))
 
-(defun jabber-read-passwd (&optional prompt)
-  "Read Jabber password, either from customized variable or from minibuffer.
-See `jabber-password'."
-  (if jabber-password
-      ;; Need to copy the password, as sasl.el wants to erase it.  The
-      ;; variable jabber-password is a high-convenience low-security
-      ;; alternative anyway.
-      (copy-sequence jabber-password)
-    (read-passwd (or prompt "Jabber password: "))))
+(defun jabber-password-key (bare-jid)
+  "Construct key for `password' library from BARE-JID."
+  (concat "xmpp:" bare-jid))
+
+(defun jabber-read-password (bare-jid)
+  "Read Jabber password from minibuffer."
+  (let ((prompt (format "Jabber password for %s: " bare-jid)))
+    (if (fboundp 'password-read)
+	;; Need to copy the password, as sasl.el wants to erase it.
+	(copy-sequence
+	 (password-read prompt (jabber-password-key bare-jid)))
+      (read-passwd prompt))))
+
+(defun jabber-cache-password (bare-jid password)
+  "Cache PASSWORD for BARE-JID."
+  (when (fboundp 'password-cache-add)
+    (password-cache-add (jabber-password-key bare-jid) password)))
+
+(defun jabber-uncache-password (bare-jid)
+  "Uncache cached password for BARE-JID.
+Useful if the password proved to be wrong."
+  (interactive (list (jabber-jid-user
+		      (completing-read "Forget password of account: " jabber-account-list))))
+  (when (fboundp 'password-cache-remove)
+    (password-cache-remove (jabber-password-key bare-jid))))
+
+(defun jabber-read-account (&optional always-ask)
+  "Ask for which connected account to use.
+If ALWAYS-ASK is nil and there is only one account, return that
+account."
+  (let ((completions
+         (mapcar (lambda (c)
+                   (cons
+                    (jabber-connection-bare-jid c)
+                    c))
+                 jabber-connections)))
+    (cond
+     ((null jabber-connections)
+      (error "Not connected to Jabber"))
+     ((and (null (cdr jabber-connections)) (not always-ask))
+      ;; only one account
+      (car jabber-connections))
+     (t
+      (or
+       ;; if there is a jabber-account property at point,
+       ;; present it as default value
+       (cdr (assoc (let ((at-point (get-text-property (point) 'jabber-account)))
+                     (when (and at-point
+                                (memq at-point jabber-connections))
+                       (jabber-connection-bare-jid at-point))) completions))
+       (let* ((default 
+                (or
+                 ;; if the buffer is associated with a connection, use it
+                 (when (and jabber-buffer-connection
+                            (memq jabber-buffer-connection jabber-connections))
+                   (jabber-connection-bare-jid jabber-buffer-connection))
+                 ;; else, use the first connection in the list
+                 (caar completions)))
+              (input (completing-read 
+                      (concat "Select Jabber account (default "
+                              default
+                              "): ")
+                      completions nil t nil nil
+                      default)))
+         (cdr (assoc input completions))))))))
 
 (defun jabber-iq-query (xml-data)
   "Return the query part of an IQ stanza.
@@ -306,9 +406,11 @@ TIME is in a format accepted by `format-time-string'."
 			 (string-to-number (substring timezone 4 6))))))))
       (encode-time second minute hour day month year timezone-seconds))))
 
-(defun jabber-report-success (xml-data context)
+(defun jabber-report-success (jc xml-data context)
   "IQ callback reporting success or failure of the operation.
-CONTEXT is a string describing the action."
+CONTEXT is a string describing the action.
+\"CONTEXT succeeded\" or \"CONTEXT failed: REASON\" is displayed in
+the echo area."
   (let ((type (jabber-xml-get-attribute xml-data 'type)))
     (message (concat context
 		     (if (string= type "result")
@@ -427,19 +529,21 @@ See secton 9.3, Stanza Errors, of XMPP Core, and JEP-0086, Legacy Errors."
    (cons 'xml-not-well-formed "XML not well formed"))
   "String descriptions of XMPP stream errors")
 
+(defun jabber-stream-error-condition (error-xml)
+  "Return the condition of a <stream:error/> tag."
+  ;; as we don't know the node name of the condition, we have to
+  ;; search for it.
+  (dolist (node (jabber-xml-node-children error-xml))
+    (when (and (string= (jabber-xml-get-attribute node 'xmlns) 
+			"urn:ietf:params:xml:ns:xmpp-streams")
+	       (assq (jabber-xml-node-name node)
+		     jabber-stream-error-messages))
+      (return (jabber-xml-node-name node)))))
+
 (defun jabber-parse-stream-error (error-xml)
   "Parse the given <stream:error/> tag and return a sting fit for human consumption."
   (let ((text-node (car (jabber-xml-get-children error-xml 'text)))
-	condition)
-    ;; as we don't know the node name of the condition, we have to
-    ;; search for it.
-    (dolist (node (jabber-xml-node-children error-xml))
-      (when (and (string= (jabber-xml-get-attribute node 'xmlns) 
-			  "urn:ietf:params:xml:ns:xmpp-streams")
-		 (assq (jabber-xml-node-name node)
-		       jabber-stream-error-messages))
-	(setq condition (jabber-xml-node-name node))
-	(return)))
+	(condition (jabber-stream-error-condition error-xml)))
     (concat (if condition (cdr (assq condition jabber-stream-error-messages))
 	      "Unknown stream error")
 	    (if (and text-node (stringp (car (jabber-xml-node-children text-node))))
@@ -507,17 +611,19 @@ See Info node `(jabber)XMPP URIs'."
     (cond
      ;; Join an MUC.
      ((string= method "join")
-      (jabber-groupchat-join jid (jabber-muc-read-my-nickname jid) t))
+      (jabber-groupchat-join
+       (jabber-read-account) jid (jabber-muc-read-my-nickname jid) t))
      ;; Register with a service.
      ((string= method "register")
-      (jabber-get-register jid))
+      (jabber-get-register (jabber-read-account) jid))
      ;; Run an ad-hoc command
      ((string= method "command")
       ;; XXX: does the 'action' attribute make sense?
-      (jabber-ahc-execute-command jid (cdr (assoc "node" args))))
+      (jabber-ahc-execute-command
+       (jabber-read-account) jid (cdr (assoc "node" args))))
      ;; Everything else: open a chat buffer.
      (t
-      (jabber-chat-with jid)))))
+      (jabber-chat-with (jabber-read-account) jid)))))
 
 (defun url-xmpp (url)
   "Handle XMPP URLs from internal Emacs functions."
@@ -525,6 +631,15 @@ See Info node `(jabber)XMPP URIs'."
   ;; url package might lose information.
   (jabber-handle-uri (url-recreate-url url)))  
 
+(defun string>-numerical (s1 s2)
+  "Return t if first arg string is more than second in numerical order."
+  (cond ((string= s1 s2) nil)
+	((> (length s1) (length s2)) t)
+	((< (length s1) (length s2)) nil)
+	((< (string-to-number (substring s1 0 1)) (string-to-number (substring s2 0 1))) nil)
+	((> (string-to-number (substring s1 0 1)) (string-to-number (substring s2 0 1))) t)
+	(t (string>-numerical (substring s1 1) (substring s2 1)))))
+
 (provide 'jabber-util)
 
 ;;; arch-tag: cfbb73ac-e2d7-4652-a08d-dc789bcded8a
