@@ -1,7 +1,7 @@
 ;; jabber.el - a minimal jabber client
 
+;; Copyright (C) 2003, 2004, 2007, 2008 - Magnus Henoch - mange@freemail.hu
 ;; Copyright (C) 2002, 2003, 2004 - tom berger - object@intelectronica.net
-;; Copyright (C) 2003, 2004 - Magnus Henoch - mange@freemail.hu
 
 ;; SSL - Support, mostly inspired by Gnus
 ;; Copyright (C) 2005 - Georg Lehner - jorge@magma.com.ni
@@ -30,27 +30,65 @@
 (defgroup jabber nil "Jabber instant messaging"
   :group 'applications)
 
-(defcustom jabber-username "emacs"
-  "jabber username (user part of JID)" 
-  :type 'string
-  :group 'jabber)
+;;;###autoload
+(defcustom jabber-account-list nil
+  "List of Jabber accounts.
+Each element of the list is a cons cell describing a Jabber account,
+where the car is a JID and the CDR is an alist.
 
-(defcustom jabber-server "magaf.org" 
-  "jabber server (domain part of JID)" 
-  :type 'string
-  :group 'jabber)
+JID is a full Jabber ID string (e.g. foo@bar.tld). You can also
+specify the resource (e.g. foo@bar.tld/emacs).
+The following keys can be present in the alist:
+:password is a string to authenticate ourself against the server.
+It can be empty.
+:network-server is a string identifying the address to connect to,
+if it's different from the server part of the JID.
+:port is the port to use (default depends on connection type).
+:connection-type is a symbol. Valid symbols are `starttls',
+`network' and `ssl'.
 
-(defcustom jabber-password nil
-  "jabber password" 
-  :type '(radio (const :tag "Prompt for password" nil)
-		 (string :tag "Save password in .emacs"))
-  :group 'jabber)
+Only JID is mandatory.  The rest can be guessed at run-time.
 
-(defcustom jabber-resource "emacs"
-  "jabber resource" 
-  :type 'string
+Examples:
+
+Two accounts without any special configuration:
+\((\"foo@example.com\") (\"bar@example.net\"))
+
+One disabled account with a non-standard port:
+\((\"romeo@montague.net\" (:port . 5242) (:disabled . t)))
+
+If you don't have SRV and STARTTLS capabilities in your Emacs,
+configure a Google Talk account like this:
+\((\"username@gmail.com\" 
+  (:network-server . \"talk.google.com\")
+  (:connection-type . ssl)))"
+  :type '(repeat
+	  (cons :tag "Account information"
+		(string :tag "JID")
+		(set :format "%v"
+		     (cons :format "%v"
+			   (const :format "" :disabled)
+			   (const :tag "Disabled" t))
+		     (cons :format "%v"
+			   (const :format "" :password)
+			   (string :tag "Password"))
+		     (cons :format "%v"
+			   (const :format "" :network-server)
+			   (string :tag "Network server"))
+		     (cons :format "%v"
+			   (const :format "" :port)
+			   (integer :tag "Port" 5222))
+		     (cons :format "%v"
+			   (const :format "" :connection-type)
+			   (choice :tag "Connection type"
+				   ;; XXX: detect whether we have STARTTLS?  option
+				   ;; for enforcing encryption?
+				   (const :tag "STARTTLS" starttls)
+				   (const :tag "Unencrypted" network)
+				   (const :tag "Legacy SSL/TLS" ssl))))))
   :group 'jabber)
 
+;;;###autoload
 (defcustom jabber-default-show ""
   "default show state"
   :type '(choice (const :tag "Online" "")
@@ -60,21 +98,18 @@
 		 (const :tag "Do not disturb" "dnd"))
   :group 'jabber)
 
+;;;###autoload
 (defcustom jabber-default-status ""
   "default status string"
   :type 'string
   :group 'jabber)
 
+;;;###autoload
 (defcustom jabber-default-priority 10
   "default priority"
   :type 'integer
   :group 'jabber)
 
-(defcustom jabber-nickname jabber-username
-  "jabber nickname, used in chat buffer prompts and as default groupchat nickname." 
-  :type 'string
-  :group 'jabber)
-
 ;;; guess internal dependencies!
 (require 'jabber-util)
 (require 'jabber-menu)
@@ -93,6 +128,7 @@
 (require 'jabber-search)
 (require 'jabber-browse)
 (require 'jabber-muc)
+(require 'jabber-muc-nick-completion)
 (require 'jabber-version)
 (require 'jabber-ahc-presence)
 (require 'jabber-modeline)
@@ -101,44 +137,37 @@
 (require 'jabber-activity)
 (require 'jabber-vcard)
 (require 'jabber-events)
+(require 'jabber-chatstates)
 (require 'jabber-vcard-avatars)
 (require 'jabber-autoaway)
 (require 'jabber-time)
+(require 'jabber-truncate)
 
 (require 'jabber-ft-client)
 (require 'jabber-ft-server)
 (require 'jabber-socks5)
 
-;; XXX: automate this some time
-(autoload 'jabber-export-roster "jabber-export"
-  "Create buffer from which roster can be exported to a file."
-  t)
-(autoload 'jabber-import-roster "jabber-export"
-  "Create buffer for roster import from FILE."
-  t)
-(autoload 'jabber-compose "jabber-compose"
-  "Create a buffer for composing a Jabber message."
-  t)
-(autoload 'jabber-private-get "jabber-private"
-  "Retrieve an item from private XML storage.
-The item to retrieve is identified by NODE-NAME (a symbol) and
-NAMESPACE (a string).
-On success, SUCCESS-CALLBACK is called with the retrieved XML fragment.
-On error, ERROR-CALLBACK is called with the entire IQ result."
-  nil)
-(autoload 'jabber-private-set "jabber-private"
-  "Store FRAGMENT in private XML storage.
-SUCCESS-CALLBACK, SUCCESS-CLOSURE-DATA, ERROR-CALLBACK and
-ERROR-CLOSURE-DATA are used as in `jabber-send-iq'."
-  nil)
+;; External notifiers
+(require 'jabber-screen)
+(require 'jabber-ratpoison)
+(require 'jabber-sawfish)
+(require 'jabber-festival)
+(require 'jabber-xmessage)
+(require 'jabber-wmii)
+(require 'jabber-osd)
 
-(defvar *jabber-current-status* ""
+(require 'jabber-autoloads)
+
+;;;###autoload
+(defvar *jabber-current-status* nil
   "the users current presence staus")
 
-(defvar *jabber-current-show* ""
+;;;###autoload
+(defvar *jabber-current-show* nil
   "the users current presence show")
 
-(defvar *jabber-current-priority* 10
+;;;###autoload
+(defvar *jabber-current-priority* nil
   "the user's current priority")
 
 (defvar *jabber-status-history* nil
@@ -165,11 +194,20 @@ ERROR-CLOSURE-DATA are used as in `jabber-send-iq'."
 (defgroup jabber-debug nil "debugging options"
   :group 'jabber)
 
+;;;###autoload
 (defcustom jabber-debug-log-xml nil
   "log all XML i/o in *-jabber-xml-log-*"
   :type 'boolean
   :group 'jabber-debug)
 
+;;;###autoload
+(defcustom jabber-debug-keep-process-buffers nil
+  "If nil, kill process buffers when the process dies.
+Contents of process buffers might be useful for debugging."
+  :type 'boolean
+  :group 'jabber-debug)
+
+;;;###autoload
 (defconst jabber-presence-faces
  '(("" . jabber-roster-user-online)
    ("away" . jabber-roster-user-away)
@@ -180,6 +218,7 @@ ERROR-CLOSURE-DATA are used as in `jabber-send-iq'."
    (nil . jabber-roster-user-offline))
  "Mapping from presence types to faces")
 
+;;;###autoload
 (defconst jabber-presence-strings
   '(("" . "Online")
     ("away" . "Away")
@@ -190,11 +229,13 @@ ERROR-CLOSURE-DATA are used as in `jabber-send-iq'."
     (nil . "Offline"))
   "Mapping from presence types to readable strings")
 
+;;;###autoload
 (defun jabber-customize ()
   "customize jabber options"
   (interactive)
   (customize-group 'jabber))
 
+;;;###autoload
 (defun jabber-info ()
   "open jabber.el manual"
   (interactive)
